@@ -32,11 +32,13 @@ use crate::error::{Error, Result};
 const BTF_SECTION_NAME: &str = ".BTF";
 
 pub(crate) struct BTF {
-    types: Vec<(u32, BtfType, *mut u8)>,
-    raw_bytes: Vec<u8>,
+    types: Vec<(u32, BtfType)>,
+    btf_hdr: btf_header,
+    raw_str_enc: Vec<u8>,
     fd: Option<RawFd>,
 }
 
+#[derive(Clone)]
 struct BtfTypeCommon {
     type_: btf_type,
     #[allow(unused)]
@@ -65,6 +67,7 @@ pub(crate) enum BtfKind {
     FloatingPoint,
 }
 
+#[derive(Clone)]
 enum BtfType {
     Integer(BtfTypeCommon, u32),
     Pointer(BtfTypeCommon),
@@ -84,6 +87,7 @@ enum BtfType {
     FloatingPoint(BtfTypeCommon),
 }
 
+#[derive(Clone)]
 struct BtfMember {
     member: btf_member,
     name: String,
@@ -119,20 +123,21 @@ impl BTF {
             return Err(Error::BTF("BTF already loaded".to_string()));
         }
 
+        let raw_bytes = self.dump()?;
         let mut v = vec![0i8; 64 * 1024];
         let log_buf = v.as_mut_ptr();
         let log_buf_size = v.capacity() * mem::size_of_val(&v[0]);
         let fd;
         unsafe {
             fd = bpf_sys::bpf_load_btf(
-                self.raw_bytes.as_ptr() as *const _,
-                self.raw_bytes.len() as u32,
-                log_buf,
+                raw_bytes.as_ptr() as *const _,
+                raw_bytes.len() as u32,
+                log_buf as _,
                 log_buf_size as u32,
                 false,
             );
             if fd < 0 {
-                let cstr = CStr::from_ptr(log_buf);
+                let cstr = CStr::from_ptr(log_buf as _);
                 error!("error on bpf_load_btf: {}", cstr.to_str().unwrap());
                 return Err(Error::IO(io::Error::last_os_error()));
             }
@@ -141,6 +146,62 @@ impl BTF {
         Ok(())
     }
 
+    fn dump(&self) -> Result<Vec<u8>> {
+        let mut raw_bytes: Vec<u8> = vec![];
+        raw_bytes.extend(unsafe {
+            slice::from_raw_parts(
+                &self.btf_hdr as *const _ as *const u8,
+                mem::size_of::<btf_header>(),
+            )
+        });
+
+        for _ in 0..self.btf_hdr.type_off {
+            raw_bytes.push(0);
+        }
+
+        for (_, type_) in self.types.iter() {
+            let offset = raw_bytes.len();
+            let sz = type_.byte_len();
+            raw_bytes.reserve(sz);
+            unsafe {
+                let out = raw_bytes.as_mut_ptr().add(offset);
+                type_.dump(out)?;
+                raw_bytes.set_len(offset + sz);
+            }
+        }
+        if raw_bytes.len()
+            != mem::size_of::<btf_header>()
+                + (self.btf_hdr.type_off + self.btf_hdr.type_len) as usize
+        {
+            error!("The length of dumped BTF types differs from the size recorded in btf_header");
+            return Err(Error::BTF(
+                "failed to dump BTF types. length mismatch".to_string(),
+            ));
+        }
+
+        if mem::size_of::<btf_header>() + (self.btf_hdr.str_off as usize) < raw_bytes.len() {
+            error!("BTF type data exceeds BTF string section offset");
+            return Err(Error::BTF("invalid BTF string section offset".to_string()));
+        }
+
+        for _ in 0..(mem::size_of::<btf_header>() + self.btf_hdr.str_off as usize - raw_bytes.len())
+        {
+            raw_bytes.push(0);
+        }
+        raw_bytes.extend(&self.raw_str_enc);
+
+        if mem::size_of::<btf_header>() + (self.btf_hdr.str_off + self.btf_hdr.str_len) as usize
+            != raw_bytes.len()
+        {
+            error!("The length of dumped BTF string section differs from the size recorded in btf_header");
+            return Err(Error::BTF(
+                "failed to dump BTF string section. length mismatch".to_string(),
+            ));
+        }
+
+        Ok(raw_bytes)
+    }
+
     fn parse_raw(bytes: &[u8]) -> Result<BTF> {
         if mem::size_of::<btf_header>() > bytes.len() {
             return Err(Error::BTF("BTF section data size is too small".to_string()));
@@ -161,11 +222,21 @@ impl BTF {
             return Err(Error::BTF("invalid binary data length".to_string()));
         }
 
-        let mut clone_bytes = bytes.to_vec();
-        let btf_types = Self::parse_types(&btf_hdr, &mut clone_bytes)?;
+        let raw_type_enc = {
+            let start = (btf_hdr.hdr_len + btf_hdr.type_off) as usize;
+            let end = start + btf_hdr.type_len as usize;
+            &bytes[start..end]
+        };
+        let mut raw_str_enc = {
+            let start = (btf_hdr.hdr_len + btf_hdr.str_off) as usize;
+            let end = start + btf_hdr.str_len as usize;
+            (&bytes[start..end]).to_vec()
+        };
+        let types = Self::parse_types(&raw_type_enc, &mut raw_str_enc)?;
         Ok(BTF {
-            types: btf_types,
-            raw_bytes: clone_bytes,
+            types,
+            btf_hdr,
+            raw_str_enc,
             fd: None,
         })
     }
@@ -181,7 +252,7 @@ impl BTF {
         let btf_bytes = &bytes[shdr.sh_offset as usize..(shdr.sh_offset + shdr.sh_size) as usize];
         let mut btf = Self::parse_raw(&btf_bytes)?;
         btf.fix_datasection(object)?;
-        for (type_id, type_, _) in btf.types.iter() {
+        for (type_id, type_) in btf.types.iter() {
             debug!("[{}] {:?}", type_id, type_);
         }
         Ok(btf)
@@ -194,14 +265,13 @@ impl BTF {
     fn fix_datasection(&mut self, object: &Elf) -> Result<()> {
         // fix size of datasection
         let mut var_type_ids = vec![];
-        for (_, type_, data) in self.types.iter_mut() {
+        for (_, type_) in self.types.iter_mut() {
             if let BtfType::DataSection(comm, vsis) = type_ {
                 let shdr = get_section_header_by_name(object, &comm.name_raw).ok_or_else(|| {
                     Error::Section(format!("DataSection not found: {}", &comm.name_raw))
                 })?;
                 comm.set_size(shdr.sh_size as u32);
                 var_type_ids.extend(vsis.iter().map(|vsi| vsi.type_));
-                type_.dump(*data)?;
             }
         }
         // fix offset of var section info
@@ -210,7 +280,7 @@ impl BTF {
             let var = self
                 .types
                 .iter()
-                .find_map(|(type_id, type_, _)| {
+                .find_map(|(type_id, type_)| {
                     if let BtfType::Variable(..) = type_ {
                         if &vti == type_id {
                             return Some(type_);
@@ -247,46 +317,31 @@ impl BTF {
                 var_offsets.insert(vti, sym.st_value);
             }
         }
-        for (_, type_, data) in self.types.iter_mut() {
+        for (_, type_) in self.types.iter_mut() {
             if let BtfType::DataSection(_, vsis) = type_ {
-                let mut modified = false;
                 for vsi in vsis.iter_mut() {
                     // offsets of variables whose linkage is static are intact
                     if let Some(offset) = var_offsets.get(&vsi.type_) {
                         vsi.offset = *offset as u32;
-                        modified = true;
                     }
                 }
-                if modified {
-                    type_.dump(*data)?;
-                }
             }
         }
         Ok(())
     }
 
     /// Helper function for parsing BPF type encoding binary data.
-    fn parse_types(
-        btf_hdr: &btf_header,
-        btf_bytes: &mut [u8],
-    ) -> Result<Vec<(u32, BtfType, *mut u8)>> {
+    fn parse_types(btf_type_enc: &[u8], btf_str_enc: &mut [u8]) -> Result<Vec<(u32, BtfType)>> {
         let mut btf_types = vec![];
-        let type_start = unsafe {
-            btf_bytes
-                .as_ptr()
-                .offset((btf_hdr.hdr_len + btf_hdr.type_off) as isize)
-        };
-        let type_end = unsafe { type_start.offset(btf_hdr.type_len as isize) };
-        let mut type_ptr = type_start;
         // type id 0 is reserved for void type. so type id is starting from 1.
         let mut type_id: u32 = 1;
-        let str_bytes = &mut btf_bytes[(btf_hdr.hdr_len + btf_hdr.str_off) as usize..];
-        while type_ptr < type_end {
-            let type_ = BtfType::parse(type_ptr as *mut _, str_bytes)?;
+        let mut remain = &btf_type_enc[..];
+        while remain.len() > 0 {
+            let type_ = BtfType::parse(remain, btf_str_enc)?;
             let sz = type_.byte_len();
-            btf_types.push((type_id, type_, type_ptr as *mut _));
-            type_ptr = unsafe { type_ptr.offset(sz as isize) };
+            btf_types.push((type_id, type_));
             type_id += 1;
+            remain = &remain[sz..];
         }
         Ok(btf_types)
     }
@@ -294,7 +349,7 @@ impl BTF {
     fn get_type_by_id(&self, type_id: u32) -> Option<&BtfType> {
         self.types
             .iter()
-            .find_map(|(tid, type_, _)| if &type_id == tid { Some(type_) } else { None })
+            .find_map(|(tid, type_)| if &type_id == tid { Some(type_) } else { None })
     }
 
     /// Get BTF type ids of a map of which symbol name is `map_sym_name`
@@ -310,7 +365,7 @@ impl BTF {
         let map_btf_type = self
             .types
             .iter()
-            .find_map(|(_, type_, _)| {
+            .find_map(|(_, type_)| {
                 if let Variable(common, _) = type_ {
                     if common.name_raw == map_btf_sym_name {
                         return Some(type_);
@@ -338,7 +393,7 @@ impl BTF {
                     .iter()
                     .find_map(|memb| {
                         if &memb.name == "key_type" {
-                            Some(memb.member.type_)
+                            Some(memb.type_id())
                         } else {
                             None
                         }
@@ -353,7 +408,7 @@ impl BTF {
                     .iter()
                     .find_map(|memb| {
                         if &memb.name == "value_type" {
-                            Some(memb.member.type_)
+                            Some(memb.type_id())
                         } else {
                             None
                         }
@@ -381,33 +436,216 @@ impl BTF {
 
     pub(crate) fn find_type_id(&self, type_name: &str, kind: BtfKind) -> Option<u32> {
         use BtfType::*;
-        self.types
-            .iter()
-            .find_map(|(type_id, type_, _)| match type_ {
-                Integer(common, _)
-                | Pointer(common)
-                | Array(common, _)
-                | Structure(common, _)
-                | Union(common, _)
-                | Enumeration(common, _)
-                | Forward(common)
-                | TypeDef(common)
-                | Volatile(common)
-                | Constant(common)
-                | Restrict(common)
-                | Function(common)
-                | FunctionProtocol(common, _)
-                | Variable(common, _)
-                | DataSection(common, _)
-                | FloatingPoint(common) => {
-                    if common.kind() == kind {
-                        if common.name_raw == type_name {
-                            return Some(*type_id);
+        self.types.iter().find_map(|(type_id, type_)| match type_ {
+            Integer(common, _)
+            | Pointer(common)
+            | Array(common, _)
+            | Structure(common, _)
+            | Union(common, _)
+            | Enumeration(common, _)
+            | Forward(common)
+            | TypeDef(common)
+            | Volatile(common)
+            | Constant(common)
+            | Restrict(common)
+            | Function(common)
+            | FunctionProtocol(common, _)
+            | Variable(common, _)
+            | DataSection(common, _)
+            | FloatingPoint(common) => {
+                if common.kind() == kind {
+                    if common.name_raw == type_name {
+                        return Some(*type_id);
+                    }
+                }
+                None
+            }
+        })
+    }
+
+    fn filter<F>(&mut self, mut f: F) -> Result<()>
+    where
+        F: FnMut(&BtfType) -> bool,
+    {
+        use BtfType::*;
+        // key: type_id / value: [type_id, type_id, ...].
+        //
+        // Value contains the types refering to the key type. Since the types
+        // in a value depend on the key type, if the key type is removed then
+        // the types the value desginates have to be eliminated first.
+        let mut in_link = RSHashMap::<u32, Vec<u32>>::with_capacity(self.types.len());
+        for (id, ty) in self.types.iter_mut() {
+            in_link.entry(*id).or_default();
+            match ty {
+                Integer(..) | Enumeration(..) | FloatingPoint(..) => {}
+                Pointer(comm)
+                | TypeDef(comm)
+                | Volatile(comm)
+                | Constant(comm)
+                | Restrict(comm)
+                | Function(comm)
+                | Variable(comm, ..)
+                | Forward(comm) => {
+                    in_link.entry(comm.type_id()).or_default().push(*id);
+                }
+                Structure(_, membs) | Union(_, membs) => {
+                    for memb in membs.iter() {
+                        in_link.entry(memb.type_id()).or_default().push(*id);
+                    }
+                }
+                FunctionProtocol(comm, params) => {
+                    let ret_type_id = comm.type_id();
+                    in_link.entry(ret_type_id).or_default().push(*id);
+                    for param in params.iter() {
+                        in_link.entry(param.type_).or_default().push(*id);
+                    }
+                }
+                DataSection(_, vsis) => {
+                    for vsi in vsis.iter() {
+                        in_link.entry(vsi.type_).or_default().push(*id);
+                    }
+                }
+                Array(_, arr) => {
+                    in_link.entry(arr.type_).or_default().push(*id);
+                    in_link.entry(arr.index_type).or_default().push(*id);
+                }
+            }
+        }
+
+        for (id, ty) in self.types.iter() {
+            if !f(ty) {
+                let mut stack = vec![];
+                let mut rem_id: u32 = *id;
+                // remove this type and all types that refer to it
+                'recur: loop {
+                    if let Some(mut in_ids) = in_link.remove(&rem_id) {
+                        if let Some(in_id) = in_ids.pop() {
+                            rem_id = in_id;
+                            stack.extend(in_ids);
+                            continue;
                         }
                     }
+
+                    if let Some(in_id) = stack.pop() {
+                        rem_id = in_id;
+                    } else {
+                        break 'recur;
+                    }
+                }
+            }
+        }
+        // type_id 0 may exist or may not exist in the in_link
+        let new_id_correc = if let Some(_) = in_link.get(&0) { 0 } else { 1 };
+
+        let mut old_ids = in_link
+            .drain()
+            .map(|(old_id, _)| old_id)
+            .collect::<Vec<u32>>();
+        old_ids.sort();
+
+        let mut new_ids = RSHashMap::<u32, u32>::with_capacity(old_ids.len());
+        for (i, old_id) in old_ids.iter().enumerate() {
+            new_ids.insert(*old_id, i as u32 + new_id_correc);
+        }
+        let mut new_types = self
+            .types
+            .iter()
+            .filter_map(|(old_id, ty)| {
+                if let Some(new_id) = new_ids.get(old_id) {
+                    Some((*new_id, (*ty).clone()))
+                } else {
                     None
                 }
             })
+            .collect::<Vec<(u32, BtfType)>>();
+
+        let mut type_len = 0;
+        for (_, ty) in new_types.iter_mut() {
+            let tydbg = format!("{:?}", ty);
+            match ty {
+                Integer(..) | Enumeration(..) | FloatingPoint(..) => {}
+                Pointer(comm)
+                | TypeDef(comm)
+                | Volatile(comm)
+                | Constant(comm)
+                | Restrict(comm)
+                | Function(comm)
+                | Variable(comm, ..)
+                | Forward(comm) => {
+                    let new_id = *new_ids.get(&comm.type_id()).ok_or_else(|| {
+                        error!("Failed to get new type id of {}", tydbg);
+                        Error::BTF("inconsistency detected in BTF".to_string())
+                    })?;
+                    comm.set_type_id(new_id);
+                }
+                Structure(_, membs) | Union(_, membs) => {
+                    for memb in membs.iter_mut() {
+                        let new_id = *new_ids.get(&memb.type_id()).ok_or_else(|| {
+                            error!(
+                                "Failed to get new type id of member {} of {}",
+                                memb.name, tydbg
+                            );
+                            Error::BTF("inconsistency detected in BTF".to_string())
+                        })?;
+                        memb.set_type_id(new_id);
+                    }
+                }
+                FunctionProtocol(comm, params) => {
+                    let new_id = *new_ids.get(&comm.type_id()).ok_or_else(|| {
+                        error!("Failed to get new ret type id of {}", tydbg);
+                        Error::BTF("inconsistency detected in BTF".to_string())
+                    })?;
+                    comm.set_type_id(new_id);
+                    for param in params.iter_mut() {
+                        let new_id = *new_ids.get(&param.type_).ok_or_else(|| {
+                            error!("Failed to get new type id of param of {}", tydbg);
+                            Error::BTF("inconsistency detected in BTF".to_string())
+                        })?;
+                        param.type_ = new_id;
+                    }
+                }
+                DataSection(_, vsis) => {
+                    for vsi in vsis.iter_mut() {
+                        let new_id = *new_ids.get(&vsi.type_).ok_or_else(|| {
+                            error!(
+                                "Failed to get new type id of variable section info of {}",
+                                tydbg
+                            );
+                            Error::BTF("inconsistency detected in BTF".to_string())
+                        })?;
+                        vsi.type_ = new_id;
+                    }
+                }
+                Array(_, arr) => {
+                    let new_id = *new_ids.get(&arr.type_).ok_or_else(|| {
+                        error!("Failed to get new type id of {}", tydbg);
+                        Error::BTF("inconsistency detected in BTF".to_string())
+                    })?;
+                    arr.type_ = new_id;
+                    let new_id = *new_ids.get(&arr.index_type).ok_or_else(|| {
+                        error!("Failed to get new index type id of {}", tydbg);
+                        Error::BTF("inconsistency detected in BTF".to_string())
+                    })?;
+                    arr.index_type = new_id;
+                }
+            }
+
+            type_len += ty.byte_len();
+        }
+
+        let decreased = self.btf_hdr.type_len - type_len as u32;
+        self.btf_hdr.type_len = type_len as u32;
+        // It is not allowed to have gap between type and str
+        // https://elixir.bootlin.com/linux/v5.13/source/kernel/bpf/btf.c#L4164
+        self.btf_hdr.str_off -= decreased;
+        // Increase str data size with zero-filled padding
+        self.btf_hdr.str_len += decreased;
+        for _ in 0..decreased {
+            self.raw_str_enc.push(0);
+        }
+        self.types = new_types;
+
+        Ok(())
     }
 }
 
@@ -450,8 +688,8 @@ impl From<u32> for BtfKind {
 }
 
 impl BtfTypeCommon {
-    fn parse(bytes: *const u8, str_bytes: &[u8]) -> Result<Self> {
-        let type_ = unsafe { ptr::read_unaligned(bytes as *const btf_type) };
+    fn parse(bytes: &[u8], str_bytes: &[u8]) -> Result<Self> {
+        let type_ = unsafe { ptr::read_unaligned(bytes.as_ptr() as *const btf_type) };
         let name = get_type_name(str_bytes, type_.name_off)?;
         Ok(Self {
             type_,
@@ -483,10 +721,14 @@ impl BtfTypeCommon {
     fn type_id(&self) -> u32 {
         unsafe { self.type_.__bindgen_anon_1.type_ }
     }
+
+    fn set_type_id(&mut self, type_: u32) {
+        self.type_.__bindgen_anon_1.type_ = type_;
+    }
 }
 
 impl BtfType {
-    fn parse(bytes: *mut u8, str_bytes: &mut [u8]) -> Result<Self> {
+    fn parse(bytes: &[u8], str_bytes: &mut [u8]) -> Result<Self> {
         let comm = BtfTypeCommon::parse(bytes, str_bytes)?;
         let vlen = comm.vlen();
         use BtfType::*;
@@ -595,7 +837,6 @@ impl BtfType {
                     );
                     comm.name_fixed = None;
                     comm.type_.name_off = 0;
-                    type_.dump(bytes)?;
                 }
             }
             Structure(comm, _)
@@ -706,40 +947,56 @@ impl BtfType {
     fn dump(&self, out: *mut u8) -> Result<()> {
         use BtfType::*;
         match self {
-            DataSection(comm, vsis) => unsafe {
-                let mut dst = out;
-                ptr::write_unaligned(dst as *mut btf_type, comm.type_);
-                dst = dst.offset(mem::size_of::<btf_type>() as isize);
-                for vsi in vsis.iter() {
-                    ptr::write_unaligned(dst as *mut btf_var_secinfo, *vsi);
-                    dst = dst.offset(mem::size_of::<btf_var_secinfo>() as isize);
-                }
-                Ok(())
-            },
-            Pointer(comm) | Volatile(comm) | Constant(comm) | Restrict(comm) => unsafe {
+            Pointer(comm) | Volatile(comm) | Constant(comm) | Restrict(comm) | Forward(comm)
+            | TypeDef(comm) | Function(comm) | FloatingPoint(comm) => unsafe {
                 ptr::write_unaligned(out as *mut btf_type, comm.type_);
-                Ok(())
             },
             Array(comm, arr) => unsafe {
                 ptr::write_unaligned(out as *mut btf_type, comm.type_);
-                ptr::write_unaligned(
-                    out.offset(mem::size_of::<btf_array>() as isize) as *mut btf_array,
-                    *arr,
-                );
-                Ok(())
+                ptr::write_unaligned(out.add(mem::size_of::<btf_array>()) as *mut btf_array, *arr);
+            },
+            Integer(comm, int) => unsafe {
+                ptr::write_unaligned(out as *mut btf_type, comm.type_);
+                ptr::write_unaligned(out.add(mem::size_of::<btf_type>()) as *mut u32, *int);
+            },
+            Variable(comm, var) => unsafe {
+                ptr::write_unaligned(out as *mut btf_type, comm.type_);
+                ptr::write_unaligned(out.add(mem::size_of::<btf_type>()) as *mut btf_var, *var);
             },
             FunctionProtocol(comm, params) => unsafe {
-                let mut dst = out;
-                ptr::write_unaligned(dst as *mut btf_type, comm.type_);
-                dst = dst.offset(mem::size_of::<btf_type>() as isize);
+                ptr::write_unaligned(out as *mut btf_type, comm.type_);
+                let mut dst = out.add(mem::size_of::<btf_type>());
                 for param in params.iter() {
                     ptr::write_unaligned(dst as *mut btf_param, *param);
-                    dst = dst.offset(mem::size_of::<btf_param>() as isize);
+                    dst = dst.add(mem::size_of::<btf_param>());
+                }
+            },
+            Structure(comm, membs) | Union(comm, membs) => unsafe {
+                ptr::write_unaligned(out as *mut btf_type, comm.type_);
+                let mut dst = out.add(mem::size_of::<btf_type>());
+                for memb in membs.iter() {
+                    ptr::write_unaligned(dst as *mut btf_member, memb.member);
+                    dst = dst.add(mem::size_of::<btf_member>());
+                }
+            },
+            Enumeration(comm, enus) => unsafe {
+                ptr::write_unaligned(out as *mut btf_type, comm.type_);
+                let mut dst = out.add(mem::size_of::<btf_type>());
+                for enu in enus.iter() {
+                    ptr::write_unaligned(dst as *mut btf_enum, *enu);
+                    dst = dst.add(mem::size_of::<btf_enum>());
+                }
+            },
+            DataSection(comm, vsis) => unsafe {
+                ptr::write_unaligned(out as *mut btf_type, comm.type_);
+                let mut dst = out.add(mem::size_of::<btf_type>());
+                for vsi in vsis.iter() {
+                    ptr::write_unaligned(dst as *mut btf_var_secinfo, *vsi);
+                    dst = dst.add(mem::size_of::<btf_var_secinfo>());
                 }
-                Ok(())
             },
-            _ => Err(Error::BTF(format!("dump is not supported for {:?}", self))),
         }
+        Ok(())
     }
 
     /// overwrite fixed name binary into string section
@@ -751,16 +1008,16 @@ impl BtfType {
     }
 
     /// `data` is a pointer to binary data of `btf_type`
-    fn read_multiple_extra<T>(data: *const u8, vlen: u32) -> Vec<T> {
-        let head_ptr = unsafe { data.offset(mem::size_of::<btf_type>() as isize) as *const T };
+    fn read_multiple_extra<T>(data: &[u8], vlen: u32) -> Vec<T> {
+        let head_ptr = unsafe { data.as_ptr().add(mem::size_of::<btf_type>()) as *const T };
         (0..vlen)
-            .map(|i| unsafe { ptr::read_unaligned(head_ptr.offset(i as isize)) })
+            .map(|i| unsafe { ptr::read_unaligned(head_ptr.add(i as usize)) })
             .collect()
     }
 
     /// `data` is a pointer to binary data of `btf_type`
-    fn read_extra<T>(data: *const u8) -> T {
-        unsafe { ptr::read_unaligned(data.offset(mem::size_of::<btf_type>() as isize) as *const T) }
+    fn read_extra<T>(data: &[u8]) -> T {
+        unsafe { ptr::read_unaligned(data.as_ptr().add(mem::size_of::<btf_type>()) as *const T) }
     }
 }
 
@@ -841,11 +1098,24 @@ impl fmt::Debug for BtfType {
                 )?;
 
                 for bm in btf_membs.iter() {
-                    write!(
-                        f,
-                        "\n\t'{}' type_id={} bits_offset={}",
-                        bm.name, bm.member.type_, bm.member.offset
-                    )?;
+                    if comm.kind_flag() {
+                        write!(
+                            f,
+                            "\n\t'{}' type_id={} bits_offset={} bitfield_size={}",
+                            bm.name,
+                            bm.type_id(),
+                            bm.bit_offset(),
+                            bm.bitfield_size(),
+                        )?;
+                    } else {
+                        write!(
+                            f,
+                            "\n\t'{}' type_id={} bits_offset={}",
+                            bm.name,
+                            bm.type_id(),
+                            bm.bit_offset()
+                        )?;
+                    }
                 }
 
                 write!(f, "")
@@ -888,7 +1158,7 @@ impl fmt::Debug for BtfType {
             FunctionProtocol(comm, ..) => {
                 write!(
                     f,
-                    "FunctionProtocol '{}' type_id={} vlen={}",
+                    "FunctionProtocol '{}' ret_type_id={} vlen={}",
                     if comm.name_raw.is_empty() {
                         &anon
                     } else {
@@ -935,6 +1205,27 @@ impl fmt::Debug for BtfType {
     }
 }
 
+impl BtfMember {
+    /// If `BtfTypeCommand::kind_flag` returns true, `self.member.offset`
+    /// contains both bit offset and bitfield size. Otherwise,
+    /// `self.member.offset` only contains bit offset.
+    fn bit_offset(&self) -> u32 {
+        self.member.offset & 0xffffff
+    }
+
+    fn bitfield_size(&self) -> u32 {
+        self.member.offset >> 24
+    }
+
+    fn type_id(&self) -> u32 {
+        self.member.type_
+    }
+
+    fn set_type_id(&mut self, type_: u32) {
+        self.member.type_ = type_;
+    }
+}
+
 fn fix_btf_identifier(btf_id: &str) -> String {
     btf_id
         .chars()
@@ -991,8 +1282,62 @@ fn get_type_name(str_bytes: &[u8], name_off: u32) -> Result<String> {
         return Err(Error::BTF("name offset is out of string data".to_string()));
     }
     Ok(unsafe {
-        CStr::from_ptr(str_bytes.as_ptr().offset(name_off as isize) as *const i8)
+        CStr::from_ptr(str_bytes.as_ptr().offset(name_off as isize) as _)
             .to_string_lossy()
             .into_owned()
     })
 }
+
+/// Fix .BTF section for `tc` command that depends on legacy BPF
+///
+/// BTF types generated by rustc contain invalid characters from the point of
+/// view of BPF verifier. Since BPF verifier only allows type identifiers and
+/// variable names compatible with C language. Since the names of BTF types
+/// that are generated by rustc also contain generic types, but they are not
+/// compatible with C language thus it is failed to load them into the kernel.
+///
+/// This function is intended to be used by `cargo-bpf` to make `tc` utility
+/// load `.BTF` section successfully. To achieve this goal, `.BTF` section is
+/// fixed during compiling the ELF object file and some BTF types that are not
+/// supported by legacy BPF of `tc` command.
+///
+/// **NOTE** Currently while compiling `tc` command, the `libbpf` can be opted
+/// in so that the `tc` command can recognize new BTF types. But most distros
+/// do not provide tc with libbpf by default but so filtering. Perhaps it is
+/// possible that `tc` compiled with `libbpf` will be installed system wide by
+/// default in the future, but it will take time. Until then it's better to fix
+/// BTF types for most users.
+pub fn tc_legacy_fix_btf_section(elf_bytes: &[u8]) -> Result<Vec<u8>> {
+    let object = Elf::parse(elf_bytes)?;
+    // let btf = BTF::parse_elf(&object, elf_bytes)?;
+
+    let shdr = get_section_header_by_name(&object, BTF_SECTION_NAME)
+        .ok_or_else(|| Error::BTF("section not found".to_string()))?;
+    let start = shdr.sh_offset as usize;
+    let end = start + shdr.sh_size as usize;
+    let raw_btf = &elf_bytes[start..end];
+    let mut btf = BTF::parse_raw(raw_btf)?;
+    btf.filter(|type_| {
+        use BtfType::*;
+        match type_ {
+            Integer(..) | Pointer(..) | Array(..) | Structure(..) | Union(..) | Enumeration(..)
+            | Forward(..) | TypeDef(..) | Volatile(..) | Constant(..) | Restrict(..)
+            | Function(..) | FunctionProtocol(..) => true,
+            _ => false,
+        }
+    })?;
+
+    let mut fixed: Vec<u8> = vec![];
+    fixed.extend(&elf_bytes[..start]);
+    fixed.extend(btf.dump()?);
+    if end < fixed.len() {
+        error!("BTF data size increased after some BTF records are filtered");
+        return Err(Error::BTF("BTF section is bloated".to_string()));
+    }
+    if end > fixed.len() {
+        error!("BTF section size decreased after some BTF records are filtered");
+        return Err(Error::BTF("BTF section is shrunk".to_string()));
+    }
+    fixed.extend(&elf_bytes[end..]);
+    Ok(fixed)
+}
